from __future__ import print_function

import string
import logging

import curses # Needed for colours

back = curses.COLOR_WHITE
front = curses.COLOR_BLACK
# Switch for white backgrounds
###back = curses.COLOR_BLACK
###front = curses.COLOR_WHITE

COLORS = [
    # Color combinations, (ID#, foreground, background)
    (1, front, back),
    (2, curses.COLOR_GREEN, -1),
    (3, curses.COLOR_BLUE, -1),
    (4, back, -1),
    (5, front, back),
    (6, curses.COLOR_CYAN, -1),
    (7, curses.COLOR_MAGENTA, -1),
    (8, curses.COLOR_BLUE, back),
    (9, curses.COLOR_GREEN, back),
    (10, curses.COLOR_RED, -1),
    (11, curses.COLOR_RED, back),
    (12, curses.COLOR_CYAN, -1),
    (13, curses.COLOR_MAGENTA, -1),
    (14, curses.COLOR_CYAN, back),
    (15, curses.COLOR_MAGENTA, back),
    (16, curses.COLOR_YELLOW, -1),
    (17, front, back),
]

name_to_color = {
    "green": 2,
    "blue": 3,
    "white": 4,
    "cyan": 6,
    "magenta": 7,
    "red": 10,
    "yellow": 16,
}

DEFAULT_COLOR = 4
OVERLAP_COLOR = 6
LINK_COLOR = 2
HELP_COLOR = 4
REF_COLOR = 3
IS_LINKED_COLOR = 16
LINE_NUMBER_COLOR = 16
SELF_LINK_COLOR = 17

COMPARE_DISAGREE_COLOR = 10
COMPARE_REF_COLOR = 12

input_action_list = {
    'leave-query-mode': [
        ('write_query', "ENTER"),
        ('write_query', '?'), ],
    'delete-query-char': [
        ('write_query', "BACKSPACE"),
        ('write_query', '!'), ], 
    'assign-text-label': [
        ('write_label', "ENTER"),
        ('write_label', '?'), ],
    'delete-label-char': [
        ('write_label', "BACKSPACE"),
        ('write_label', '!'), ], 
    'clear-query': [
        '|', ],
    'enter-query-mode': [
        '\\', ],
    'enter-label-mode': [
        't', ],
    'toggle-line-numbers': [
        '#', ],
    'move-up': [
        "UP", 'i', ],
    'move-down': [
        "DOWN", 'o', ],
    'move-left': [
        "LEFT", 'j', ],
    'move-right': [
        "RIGHT", ';', ],
    'jump-up': [
        "SHIFT-UP", 'I', ],
    'jump-down': [
        "SHIFT-DOWN", 'O', ],
    'jump-left': [
        "SHIFT-LEFT", 'J', ],
    'jump-right': [
        "SHIFT-RIGHT", ':', ],
    'move-link-up': [
        ('link', "SHIFT-UP"), ('link', 'I'), ],
    'move-link-down': [
        ('link', "SHIFT-DOWN"), ('link', 'O'), ],
    'move-link-left': [
        ('link', "SHIFT-LEFT"), ('link', 'J'), ],
    'move-link-right': [
        ('link', "SHIFT-RIGHT"), ('link', ':'), ],
    'page-up': [
        "PPAGE", '{' ],
    'page-down': [
        "NPAGE", '}' ],
    'extend-up': [
         ],
    'extend-down': [
        ],
    'contract-up': [
         ],
    'contract-down': [
         ],
    'extend-left': [
        'm', ],
    'extend-right': [
        '/', ],
    'contract-left': [
        'k', ],
    'contract-right': [
        'l', ],
    'extend-link-left': [
        'M', ],
    'extend-link-right': [
        '?', ],
    'contract-link-left': [
        'K', ],
    'contract-link-right': [
        'L', ],
    'toggle-help': [
        'h', ],
    'next-file': [
        ']', ('no_file', ']'), ],
    'previous-file': [
        '[', ('no_file', '['), ],
    'quit': [
        'Q', ],
    'save-and-quit': [
        'q', ],
    'save': [
        's', ],
    'create-link': [
        ('link', 'D'), ],
    'create-link-and-move': [
        ('link', 'd'), ],
    'edit-annotation': [
        ],
    'remove-annotation': [
        (None, 'u'), ],
    'update-num': [
        '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', ],
    'toggle-progress': [
            ['>', 'p'], ],
    'toggle-legend': [
            ['>', 'l'], ],
    'toggle-current-mark': [
            ['>', 'm'], ],
    'search-previous': [
        'p', ],
    'search-next': [
        'n', ],
    'search-link-previous': [
        'P', ],
    'search-link-next': [
        'N', ],
}

# Fill in all other characters for searching
used = set()
for key in input_action_list:
    for item in input_action_list[key]:
        if type(item) == tuple:
            used.add((item[0], tuple(item[1])))
        else:
            used.add((None, tuple(item)))
###for item in used:
###    print("Used", item)
for char in string.printable:
    if char in string.whitespace:
        if char == ' ':
            char = 'SPACE'
        else:
            continue
    if ('write_query', (char,)) not in used:
        input_action_list.setdefault('add-to-query', []).append(('write_query', char))
    if ('write_label', (char,)) not in used:
        input_action_list.setdefault('add-to-label', []).append(('write_label', char))

key_to_symbol = {
    curses.KEY_UP: "UP",
    curses.KEY_DOWN: "DOWN",
    curses.KEY_LEFT: "LEFT",
    curses.KEY_RIGHT: "RIGHT",
    curses.KEY_SLEFT: "SHIFT-LEFT",
    curses.KEY_SRIGHT: "SHIFT-RIGHT",
    curses.KEY_NPAGE: "NPAGE",
    curses.KEY_PPAGE: "PPAGE",
}
for c in string.printable:
    if c not in string.whitespace:
        key_to_symbol[ord(c)] = c
    elif c == ' ':
        key_to_symbol[ord(c)] = 'SPACE'
    elif c == '\t':
        key_to_symbol[ord(c)] = 'TAB'
    elif c == '\n':
        key_to_symbol[ord(c)] = 'ENTER'
special_keys = { # OS X
    337: "SHIFT-UP",
    336: "SHIFT-DOWN",
    10: "ENTER",
    127: "BACKSPACE",
}
for key in special_keys:
    key_to_symbol[key] = special_keys[key]
symbol_to_key = {}
for key in key_to_symbol:
    symbol_to_key[key_to_symbol[key]] = key

def keydef_to_symbols(keydef):
    symbols = ['']
    for char in keydef:
        if char == '_' and len(symbols[-1]) > 0:
            symbols.append('')
        else:
            symbols[-1] += char
    return symbols

class Config(object):
    def __init__(self, args, labels={}):
        self.args = args
        self.labels = labels
        self.annotation = args.ann_scope
        self.annotation_type = args.ann_type
        self.input_to_action = {}

        if args.config_file is not None:
            for line in open(args.config_file):
                # Set general command keybindings
                if line.startswith("Input:"):
                    _, action, mode, key = line.strip().split()
                    symbols = tuple(keydef_to_symbols(key))
                    if mode == 'all':
                        mode = None
                    self.add_keybinding(mode, symbols, action)
                elif line.startswith("Label:"):
                    _, label, key, color = line.strip().split()
                    symbols = tuple(keydef_to_symbols(key))
                    self.labels[label] = (symbols, color)
                elif line.startswith("Special_Key:"):
                    _, symbol, key = line.strip().split()
                    special_keys[int(key)] = symbol
                    symbol_to_key[symbol] = key
                    key_to_symbol[key] = symbol
        else:
            # Core key set as defined above
            for action in input_action_list:
                for opt in input_action_list[action]:
                    mode, symbol = None, opt
                    if type(opt) == tuple:
                        mode = opt[0]
                        symbol = opt[1]
                    if type(symbol) == str:
                        symbol = [symbol]
                    self.add_keybinding(mode, tuple(symbol), action)
            # Provided labels
            for label in self.labels:
                key, _ = self.labels[label]
                self.add_keybinding('category', key, 'edit-annotation')

        # Fill annotation keys
        self.input_to_label = {}
        for label in self.labels:
            key, _ = self.labels[label]
            if type(key) == str:
                key = (key,)
            else:
                key = tuple(key)
            self.input_to_label[key] = label

        self.valid_prefixes = set()
        for mode, symbol in self.input_to_action:
            for i in range(1, len(symbol)):
                self.valid_prefixes.add((mode, symbol[:i]))
        for mode, symbol in self.input_to_action:
            if (mode, symbol) in self.valid_prefixes:
                raise Exception("input {} overlaps with a prefix".format(symbol))

    def get_color_for_label(self, mark):
        name = self.labels[mark][1]
        return name_to_color[name]

    def get_label_for_input(self, user_input):
        return self.input_to_label.get(user_input, None)

    def add_keybinding(self, mode, key, action):
        pair = (mode, key)
        if pair in self.input_to_action:
            raise Exception("input {} used twice".format(pair))
        self.input_to_action[pair] = action

    def __str__(self):
        ans = []

        for mode, key in self.input_to_action:
            action = self.input_to_action[mode, key]
            if mode is None:
                mode = 'all'
            key = "_".join(key)

            ans.append("{:<15} {:<25} {:<20} {}".format("Input:", action, mode, key))

        for label in self.labels:
            key, color = self.labels[label]
            key = "_".join(key)
            ans.append("{:<15} {:<25} {} {}".format("Label:", label, key, color))

        for key in special_keys:
            symbol = special_keys[key]
            ans.append("{:<15} {:<25} {}".format("Special_Key:", symbol, key))

        return "\n".join(ans)




from __future__ import print_function

import logging
import glob
import sys

from .config import *

def process_fileinfo(file_info, config):
    filenames = []
    for line in file_info:
        # A line describing a file in the form:
        #  raw_file [output_file [cur_line cur_token [other_annotations]]]
        parts = line.split()
        next_part = 0

        # Input data
        raw_file = parts[0]
        next_part += 1

        # Output file name
        output_file = raw_file + ".annotations"
        if len(parts) > next_part:
            output_file = parts[next_part]
            next_part += 1

        # Start somewhere other than the top
        d = Document(raw_file)
        position = Span(config.annotation, d)
        if len(parts) > next_part:
            position_text = []
            depth = 0
            first = True
            while depth > 0 or first:
                for char in parts[next_part]:
                    if char == '(': depth += 1
                    if char == ')': depth -= 1
                position_text.append(parts[next_part])
                first = False
                next_part += 1
            position_text = ' '.join(position_text)

            span = eval(position_text)
            position = Span(config.annotation, d, span)

        # Additional annotations (used when comparing annotations)
        annotations = []
        if len(parts) > next_part:
            annotations = parts[next_part:]
        filenames.append((raw_file, position, output_file, annotations))

    # Check files exist (or do not exist if being created)
    missing = []
    extra = []
    for raw_file, _, output_file, annotations in filenames:
        if len(glob.glob(raw_file)) == 0:
            missing.append(raw_file)
        if not config.args.overwrite:
            if len(glob.glob(output_file)) != 0:
                extra.append(output_file)
        for annotation in annotations:
            if len(glob.glob(annotation)) == 0:
                missing.append(annotation)
    if len(missing) > 0 or len(extra) > 0:
        error = "Input Filename List Has Errors"
        if len(missing) > 0:
            error += "\n\nUnable to open:\n" + '\n'.join(missing)
        if len(extra) > 0:
            error += "\n\nOutput file already exists:\n" + '\n'.join(extra)
        raise Exception(error)
    return filenames

class Document(object):
    """Storage for the raw text data."""
    # TODO: Think about maintaining whitespace variations, probably just by
    # removing the .strip() below and then adjusting token selection to skip
    # blank tokens

    def __init__(self, filename):
        self.raw_text = open(filename).read()
        self.lines = self.raw_text.split("\n")
        self.search_cache = {}

        self.tokens = []
        self.first_char = None
        self.last_char = None
        for line in self.raw_text.split("\n"):
            cur = []
            self.tokens.append(cur)
            for token in line.strip().split():
                if self.first_char is None:
                    self.first_char = (len(self.tokens) - 1, 0, 0)
                self.last_char = (len(self.tokens) - 1, len(cur), len(token) - 1)
                cur.append(token)
        assert self.first_char is not None, "Empty document"

    def get_3tuple(self, partial, start):
        if len(partial) == 3:
            return partial
        elif len(partial) == 0:
            if start: return self.first_char
            else: return self.last_char
        else:
            line = partial[0]

            token = 0
            if len(partial) == 1 and (not start):
                token = len(self.tokens[line]) - 1
            elif len(partial) == 2:
                token = partial[1]

            char = 0
            if not start:
                char = len(self.tokens[line][token]) - 1

            return (line, token, char)

    def matches(self, text):
        if text not in self.search_cache:
            positions = []
            self.search_cache[text] = positions
            for line_no, line in enumerate(self.lines):
                if text in line:
                    parts = line.split(text)
                    ctoken, cchar = 0, 0
                    options = []
                    for part in parts[:-1]:
                        # Advance
                        for char in part:
                            if char == ' ':
                                ctoken += 1
                                cchar = 0
                            else:
                                cchar += 1
                        positions.append((line_no, ctoken, cchar))
                        for char in text:
                            if char == ' ':
                                ctoken += 1
                                cchar = 0
                            else:
                                cchar += 1
        return self.search_cache[text]

    def get_moved_pos(self, pos, right=0, down=0, maxjump=False, skip_blank=True):
        """Calculate a shifted version of  a given a position in this document.

        Co-ordinates are (line number, token number, character number), with
        (0,0, 0) as the top left, tokens increasing left to right and lines
        increasing top to bottom.
        """
        if len(pos) == 0: # This is the whole document, can't move
            return pos
        elif len(pos) == 1: # This is a line
            npos = pos[0]
            # Interpret left/right as also being up/down for lines
            if down == 0 and right != 0:
                down = right
            if maxjump:
                if down < 0: npos = self.first_char[0]
                elif down > 0: npos = self.last_char[0]
            else:
                # Shift incrementally so we can optionally only count lines
                # that have tokens.
                shift = down
                delta = 1 if shift > 0 else -1
                while shift != 0 and self.first_char[0] <= npos + delta <= self.last_char[0]:
                    npos += delta
                    if (not skip_blank) or len(self.tokens[npos]) > 0:
                        shift -= delta

            return (npos,)
        elif len(pos) == 2: # Moving a token
            nline = pos[0]
            ntok = pos[1]

            # Vertical movement
            if maxjump:
                # We always want to be on a token, so go to the first or last
                # line with one.
                if down < 0: nline = self.first_char[0]
                elif down > 0: nline = self.last_char[0]
            else:
                # Shift incrementally so we can optionally only count lines
                # that have tokens.
                shift = down
                delta = 1 if shift > 0 else -1
                while shift != 0 and self.first_char[0] <= nline + delta <= self.last_char[0]:
                    nline += delta
                    if (not skip_blank) or len(self.tokens[nline]) > 0:
                        shift -= delta

            # Horizontal movement
            ntok = min(ntok, len(self.tokens[nline]) - 1)
            if maxjump:
                if right < 0: ntok = 0
                elif right > 0: ntok = len(self.tokens[nline]) - 1
            else:
                shift = right
                delta = 1 if shift > 0 else -1
                while shift != 0:
                    if delta == -1 and nline == self.first_char[0] and ntok == self.first_char[1]:
                        break
                    if delta == 1 and nline == self.last_char[0] and ntok == self.last_char[1]:
                        break
                    if 0 <= ntok + delta < len(self.tokens[nline]):
                        ntok += delta
                    else:
                        # Go forward/back to a line with tokens. Note, we know
                        # there are later/earlier lines, since otherwise we
                        # would have been at the last_char/first_char
                        # position.
                        nline += delta
                        while len(self.tokens[nline]) == 0:
                            nline += delta
                        ntok = 0 if delta > 0 else len(self.tokens[nline]) - 1
                    shift -= delta
            return (nline, ntok)
        else: # Moving a character
            # Vertical movement
            nline = pos[0]
            if maxjump:
                # We always want to be on a character, so go to the first or
                # last line with one.
                if down < 0: nline = self.first_char[0]
                elif down > 0: nline = self.last_char[0]
            else:
                # Shift incrementally because we only want to count lines that
                # have characters.
                shift = down
                delta = 1 if shift > 0 else -1
                while shift != 0 and self.first_char[0] <= nline + delta <= self.last_char[0]:
                    nline += delta
                    if (not skip_blank) or len(self.tokens[nline]) > 0:
                        shift -= delta

            # Horizontal movement
            ntok = min(len(self.tokens[nline]) - 1, pos[1])
            nchar = min(len(self.tokens[nline][ntok]) - 1, pos[2])
            if maxjump:
                if right < 0:
                    ntok = 0
                    nchar = 0
                elif right > 0:
                    ntok = len(self.tokens[nline]) - 1
                    nchar = len(self.tokens[nline][ntok]) - 1
            else:
                shift = right
                delta = 1 if shift > 0 else -1
                while shift != 0:
                    if delta == -1 and \
                            nline == self.first_char[0] and \
                            ntok == self.first_char[1] and \
                            nchar == self.first_char[1]:
                        break
                    if delta == 1 and \
                            nline == self.last_char[0] and \
                            ntok == self.last_char[1] and \
                            nchar == self.last_char[1]:
                        break
                    if 0 <= nchar + delta < len(self.tokens[nline][ntok]):
                        nchar += delta
                    elif delta < 0 and ntok > 0:
                        ntok -= 1
                        nchar = len(self.tokens[nline][ntok]) - 1
                    elif delta > 0 and ntok < len(self.tokens[nline]) - 1:
                        ntok += 1
                        nchar = 0
                    else:
                        # Go forward/back to a line with tokens. Note, we know
                        # there are later/earlier lines, since otherwise we
                        # would have been at the last_char/first_char
                        # position.
                        if nline + delta <= self.last_char[0]:
                            nline += delta
                            while len(self.tokens[nline]) == 0:
                                nline += delta
                            ntok = 0 if delta > 0 else len(self.tokens[nline]) - 1
                            nchar = 0 if delta > 0 else len(self.tokens[nline][ntok]) - 1
                    shift -= delta
            return (nline, ntok, nchar)

    def get_next_pos(self, pos):
        if len(pos) == 0:
            return pos
        elif len(pos) == 1:
            return self.get_moved_pos(pos, 0, 1)
        else:
            return self.get_moved_pos(pos, 1, 0)

    def get_previous_pos(self, pos):
        if len(pos) == 0:
            return pos
        elif len(pos) == 1:
            return self.get_moved_pos(pos, 0, -1)
        else:
            return self.get_moved_pos(pos, -1, 0)

###class SpanCompare(Enum):
###  smaller = 0
###  smaller_left = 1
###  overlap_end = 2
###  overlap_right = 3
###  cover = 4
###  left_inside = 5
###  equal = 6
###  left_overlap = 7
###  inside = 8
###  inside_right = 9
###  overlap_start = 10
###  right_larger = 11
###  larger = 12
###  one_smaller = 13
###  one_left = 14
###  one_inside = 15
###  one_right = 16
###  one_larger = 17
###  smaller_one = 18
###  smaller_match = 19
###  cover_one = 20
###  equal_one = 21
###  match_larger = 22
###  larger_one = 23
###  smaller_one_one = 24
###  larger_one_one = 25

value_from_comparisons = {
#  s0s1 e0e1 s0e1 e0s1 s0e0 s1e1 SpanCompare                |-------|
  (1, 1, 1, 1, 1, 1): "smaller",          # .--.                 
  (1, 1, 1, 0, 1, 1): "smaller_left",     # .-----|              
  (1, 1, 1, -1, 1, 1): "overlap_end",     # .---------.          
  (1, 0, 1, -1, 1, 1): "overlap_right",   # .-------------|      
  (1, -1, 1, -1, 1, 1): "cover",          # .-------------------.
  (0, 1, 1, -1, 1, 1): "left_inside",     #       |---.          
  (0, 0, 1, -1, 1, 1): "equal",           #       |-------|      
  (0, -1, 1, -1, 1, 1): "left_overlap",   #       |-------------.
  (-1, 1, 1, -1, 1, 1): "inside",         #          .-.         
  (-1, 0, 1, -1, 1, 1): "inside_right",   #           .---|      
  (-1, -1, 1, -1, 1, 1): "overlap_start", #           .---------.
  (-1, -1, 0, -1, 1, 1): "right_larger",  #               |-----.
  (-1, -1, -1, -1, 1, 1): "larger",       #                  .--.
  (1, 1, 1, 1, 0, 1): "one_smaller",      # .                    
  (0, 1, 1, 0, 0, 1): "one_left",         #       |              
  (-1, 1, 1, -1, 0, 1): "one_inside",     #           .          
  (-1, 0, 0, -1, 0, 1): "one_right",      #               |      
  (-1, -1, -1, -1, 0, 1): "one_larger",   #                     .
  # Now consider cases where the second has width 0
  #                                                             |          
  (1, 1, 1, 1, 1, 0): "smaller_one",      #     .--.             
  (1, 0, 1, 0, 1, 0): "smaller_match",    #     .-----|          
  (1, -1, 1, -1, 1, 0): "cover_one",      #     .-----------.    
  (0, 0, 0, 0, 0, 0): "equal_one",        #           |          
  (0, -1, 0, -1, 1, 0): "match_larger",   #           |-----.    
  (-1, -1, -1, -1, 1, 0): "larger_one",   #              .--.    
  (1, 1, 1, 1, 0, 0): "smaller_one_one",  #     .                
  (-1, -1, -1, -1, 0, 0): "larger_one_one", #               .    
}

span_compare_ge = {
    "left_inside", "equal", "left_overlap",
    "inside", "inside_right", "overlap_start",
    "right_larger", "larger", "one_left",
    "one_inside", "one_right", "one_larger",
    "equal_one", "match_larger", "larger_one",
    "larger_one_one"
}
span_compare_le = {
    "smaller", "smaller_left", "overlap_end",
    "overlap_right", "left_inside", "equal",
    "inside", "inside_right", "one_smaller",
    "one_left", "one_inside", "one_right",
    "smaller_one", "smaller_match", "equal_one",
    "smaller_one_one"
}

class Span(object):
    """A continuous span of text.
    
    All annotations are on spans, some of which just happen to have a single element."""

    def __init__(self, scope, doc, span=None):
        self.start = None
        self.end = None
        self.doc = doc
        self.scope = scope

        # Most of the time a span will be provided to start from.
        if span is None:
            first = self.doc.first_char
            if scope == 'character':
                self.start = (first[0], first[1], first[2])
            elif scope == 'token':
                self.start = (first[0], first[1])
            elif scope == 'line':
                self.start = (first[0],)
            elif scope == 'document':
                self.start = ()
            else:
                raise Exception("Invalid scope")
            self.end = self.start
        else:
            # Check it has the right length
            length = None
            if scope == 'character': length = 3
            elif scope == 'token': length = 2
            elif scope == 'line': length = 1
            elif scope == 'document': length = 0
            else: raise Exception("Invalid scope")

            # TODO: Add a check that this position is valid for this doc
            if type(span) == int and length == 1:
                self.start = (span,)
                self.end = (span,)
            elif type(span) == tuple:
                if len(span) == 0:
                    self.start, self.end = (), ()
                else:
                    if type(span[0]) == int:
                        assert len(span) == length, "Invalid item: got {} not {}".format(len(span), length)
                        self.start = span
                        self.end = span
                    else:
                        assert len(span[0]) == len(span[1]) == length, "Invalid item: got {} not {}".format(len(span), length)
                        self.start = span[0]
                        self.end = span[1]
            else:
                assert len(span.start) == len(span.end) == length, "Invalid item: got {} not {}".format(len(span), length)
                self.start = span.start
                self.end = span.end

    def _compare_tuples(self, a, b):
        # Returns a number that is the kind of delta going from a to b
        # (positive, negative, or zero)
        if len(a) == 0 or len(b) == 0:
            return 0
        if a[0] == b[0]:
            if len(a) == 1 or len(b) == 1:
                return 0
            if a[1] == b[1]:
                if len(a) == 2 or len(b) == 2:
                    return 0
                if a[2] == b[2]:
                    return 0
                elif a[2] < b[2]:
                    return 1
                else:
                    return -1
            elif a[1] < b[1]:
                return 1
            else:
                return -1
        elif a[0] < b[0]:
            return 1
        else:
            return -1

    def __hash__(self):
        return hash((self.start, self.end))
    def __eq__(self, other):
        if type(self) != type(other):
            return False
        elif self._compare_tuples(self.start, other.start) != 0:
            return False
        elif self._compare_tuples(self.end, other.end) != 0:
            return False
        else:
            return True
    def __lt__(self, other):
        assert type(self) == type(other)
        comp = self._compare_tuples(self.start, other.start)
        if comp == 0:
            return self._compare_tuples(self.end, other.end) == 1
        return comp == 1

    def __ne__(self, other):
        return not self.__eq__(other)
    def __le__(self, other):
        return self.__lt__(other) or self.__eq__(other)
    def __gt__(self, other):
        return not self.__le__(other)
    def __ge__(self, other):
        return not self.__lt__(other)

    def __repr__(self):
        return "Span({}, {})".format(self.start, self.end)

    def __str__(self):
        return str((self.start, self.end))

    def compare(self, other):
        '''Compares two spans and returns a SpanCompare.'''
        assert type(self) == type(other)

        s0 = self.doc.get_3tuple(self.start, True)
        e0 = self.doc.get_3tuple(self.end, False)
        s1 = other.doc.get_3tuple(other.start, True)
        e1 = other.doc.get_3tuple(other.end, False)

        s0s1 = self._compare_tuples(s0, s1)
        s0e1 = self._compare_tuples(s0, e1)
        e0s1 = self._compare_tuples(e0, s1)
        e0e1 = self._compare_tuples(e0, e1)
        s0e0 = self._compare_tuples(s0, e0)
        s1e1 = self._compare_tuples(s1, e1)

        return value_from_comparisons[s0s1, e0e1, s0e1, e0s1, s0e0, s1e1]

    def to_3tuple(self):
        if self.scope == 'character':
            return self
        start = self.doc.get_3tuple(start)
        end = self.doc.get_3tuple(end)
        return Span('character', self.doc, (start, end))

    def search(self, query, direction=None, count=1, maxjump=False):
        options = self.doc.matches(query)
        logging.debug(options)
        ans = None
        for option in options:
            comp = self._compare_tuples(self.start, option)
            if comp < 0 and direction == 'previous':
                ans = option
            elif comp > 0:
                if direction == 'next':
                    ans = option
                break
        if ans is None:
            return self
        else:
            return Span(self.scope, self.doc, ans[:len(self.start)])

    def edited(self, direction=None, change=None, distance=1, maxjump=False):
        """Change this span, either moving both ends or only one.

        direction is left, right, up, down, next or previous
        change is move, expand, or contract
        distance is an integer, with negative numbers meaning max
        """
###        logging.debug("{} {} {} {}".format(self, direction, change, distance))
        new_start = self.start
        new_end = self.end

        if direction == 'next':
            new_start = self.doc.get_next_pos(self.start)
            new_end = self.doc.get_next_pos(self.end)
            return Span(self.scope, self.doc, (new_start, new_end))
        elif direction == 'previous':
            new_start = self.doc.get_previous_pos(self.start)
            new_end = self.doc.get_previous_pos(self.end)
            return Span(self.scope, self.doc, (new_start, new_end))

        right = 0
        down = 0
        if direction == 'left':
            right = -distance
        elif direction == 'right':
            right = distance
        elif direction == 'up':
            down = -distance
        elif direction == 'down':
            down = distance

        if change == 'contract':
            right *= -distance
            down *= -distance

        if change == "move":
            nstart = self.doc.get_moved_pos(new_start, right, down, maxjump)
            nend = self.doc.get_moved_pos(new_end, right, down, maxjump)
###            logging.debug("From {} and {} to {} and {}".format(self.start, self.end, nstart, nend))
            # Only move if it will change both (otherwise it is a shift).
            if nstart != self.start and nend != self.end:
                new_start = nstart
                new_end = nend
        else:
###            logging.debug("From {} do {} {} {} {} {} {}".format(self, direction, change, distance, maxjump, right, down))
            move_start = direction == "left" or direction == "up"

            to_move = new_end
            if move_start:
                to_move = new_start
            moved = self.doc.get_moved_pos(to_move, right, down, maxjump)

            nstart = new_start
            nend = new_end
            if move_start: nstart = moved
            else: nend = moved

            if self._compare_tuples(nstart, nend) >= 0:
                new_start = nstart
                new_end = nend

        ans = Span(self.scope, self.doc, (new_start, new_end))
###        logging.debug("Returning {}".format(ans))
        return ans
    
    # How to do coreference resolution annotation:
    # - Normal mode is selecting a position using the edit function
    # - Switch to link mode and then toggle between mentions including this one (to indicate no prior link)

class Item(object):
    """One or more spans and a set of labels.

    This is used in Datum to keep track of annotations, and used in View to determine the current appearance."""
    def __init__(self, doc, init_span=None, init_label=None):
        self.doc = doc
        self.spans = []
        if type(init_span) == list:
            self.spans += init_span
        elif init_span is not None:
            self.spans.append(init_span)

        self.labels = set()
        if type(init_label) == set:
            self.labels.update(init_label)
        elif init_label is not None:
            self.labels.add(init_label)

    def __eq__(self, other):
        return self.spans == other.spans and self.labels == other.labels and self.doc == other.doc

    def __str__(self):
        labels = []
        for label in self.labels:
            labels.append(str(label))
        labels = ' '.join(labels)

        spans = '[' + ', '.join([str(s) for s in self.spans]) +']'
        if len(self.spans) == 1:
            spans = str(self.spans[0])
            if self.spans[0].start == self.spans[0].end:
                spans = str(self.spans[0].start)
                if len(self.spans[0].start) == 1:
                    spans = str(self.spans[0].start[0])
        elif len(self.spans) > 1:
            all_single = True
            for s in self.spans:
                if s.start != s.end:
                    all_single = False
            if all_single:
                spans = str([s.start for s in self.spans])
                if len(self.spans[0].start) == 1:
                    spans = " ".join([str(s.start[0]) for s in self.spans])

        return "{} - {}".format(spans, labels)

def get_spans(text, doc, config):
    # TODO: allow for <filename>:data

    spans = []
    if text[0] in '[(':
        spans = eval(text.strip())
        if type(spans) == int:
            spans = [(spans,)]
        elif type(spans) == tuple:
            spans = [spans]
        elif type(spans) == list:
            if len(spans) == 0:
                spans = [()]
            elif type(spans[0]) == int:
                spans = [(s,) for s in spans]
    else:
        for num in text.split():
            spans.append((int(num),))
    
    return [Span(config.annotation, doc, s) for s in spans]

def get_labels(text, config):
    labels = set()
    if config.annotation_type == 'categorical':
        for label in text.strip().split():
            labels.add(label)
    else:
        assert len(text.strip().split()) == 0, text

    return labels

def read_annotation_file(config, filename, doc):
    items = []
    if len(glob.glob(filename)) == 1:
        for line in open(filename):
            # Each line is:
            # [spans] - [labels]
            fields = line.strip().split()
            spans = get_spans(line.split('-')[0], doc, config)
            labels = get_labels('-'.join(line.split('-')[1:]), config)
            items.append(Item(doc, spans, labels))
        logging.info("Read {}".format(filename))

    return items

class Datum(object):
    """Storage for a single file's data and annotations.

    Note, the structure of storage depends on the annotation type."""

    def __init__(self, filename, config, output_file, other_annotation_files):
        self.filename = filename
        self.config = config
        self.output_file = output_file
        self.doc = Document(filename)
        logging.info("Reading data from "+ self.output_file)
        self.annotations = read_annotation_file(config, self.output_file, self.doc)

        self.other_annotation_files = other_annotation_files
        self.other_annotations = []
        for filename in other_annotation_files:
            self.other_annotations.append(read_annotation_file(config, filename, self.doc))

        # Working this out is a once-off expensive process
        self.disagreements = []
        all_item_counts = {}
        hash_to_item = {}
        for annotations in self.other_annotations:
            for item in annotations:
                h = hash((tuple(item.spans), tuple(item.labels)))
                hash_to_item[h] = item
                if h not in all_item_counts:
                    all_item_counts[h] = 0
                all_item_counts[h] += 1
        for h, count in all_item_counts.items():
            self.disagreements.append((hash_to_item[h],len(self.other_annotations) - count))

    def get_next_self_link(self, cursor, linking_pos, direction, moving_link):
        if moving_link:
            self_links = set()
            for item in self.annotations:
                if max(item.spans) == min(item.spans):
                    self_links.add(min(item.spans))
            position = linking_pos.edited(direction)
            prev = None
            while position != prev:
                if position in self_links:
                    return position
                prev = position
                position = position.edited(direction)
            return linking_pos
        else:
            return cursor

    def get_next_unannotated(self, cursor, linking_pos, direction, moving_link):
        if moving_link:
            annotated = set()
            for item in self.annotations:
                span = max(item.spans)
                annotated.add(span)
            position = linking_pos.edited(direction)
            prev = None
            while position in annotated and position != prev:
                prev = position
                position = position.edited(direction)
            if position not in annotated:
                return position
            return linking_pos
        else:
            return cursor

    def get_next_disagreement(self, cursor, linking_pos, direction, moving_link, cycle=True):
        best = None
        first = None
        last = None
        for item, count in self.disagreements:
            if moving_link:
                if count > 0:
                    span = max(item.spans)
                    if first is None or span > first:
                        first = span
                    if last is None or span < last:
                        last = span
                    if direction == 'next' and span > linking_pos:
                        if best is None or span < best:
                            best = span
                    elif direction == 'previous' and span < linking_pos:
                        if best is None or span > best:
                            best = span
            else:
                has_link = False
                for span in item.spans:
                    if span == linking_pos:
                        has_link = True
                if has_link:
                    for span in item.spans:
                        if span <= linking_pos:
                            if first is None or span < first:
                                first = span
                            if last is None or span > last:
                                last = span
                            if direction == 'next' and span > cursor:
                                if best is None or span < best:
                                    best = span
                            elif direction == 'previous' and span < cursor:
                                if best is None or span > best:
                                    best = span
        if best is None and cycle:
            if direction == 'next':
                best = first
            else:
                best = last
        return best

    def get_all_markings(self, cursor, linking_pos):
        ans = {}

        # Set colors for cursor and linking pos
        pos = cursor.start
        while True:
            ans.setdefault(pos, []).append('cursor')
            if pos == cursor.end:
                break
            pos = self.doc.get_next_pos(pos)
            # Handle the case of a space
            if len(pos) == 2 or (len(pos) == 3 and pos[2] == 0):
                ans.setdefault((pos[0], pos[1], -1), []).append('cursor')
        if linking_pos is not None:
            pos = linking_pos.start
            while True:
                ans.setdefault(pos, []).append('link')
                if pos == linking_pos.end:
                    break
                pos = self.doc.get_next_pos(pos)
                # Handle the case of a space
                if len(pos) == 2 or (len(pos) == 3 and pos[2] == 0):
                    ans.setdefault((pos[0], pos[1], -1), []).append('link')

        # Set item colors
        for item in self.annotations:
            # Get the standard color for this item based on its label
            base_labels = []
            if self.config.annotation_type == 'categorical':
                # For categorical use the configuration set
                for key in item.labels:
                    if key in self.config.labels:
                        base_labels.append(key)
                    else:
                        base_labels.append("label:"+ key)
            elif self.config.annotation_type == 'link':
                # For links potentially indicate it is linked
                if not self.config.args.do_not_show_linked:
                    base_labels.append('linked')
            
            is_self_link = len(item.spans) == 2 and item.spans[0] == item.spans[1]

            has_link = False
            for span in item.spans:
                if span == linking_pos:
                    has_link = True

            for span in item.spans:
                pos = span.start
                while True:
                    cur = ans.setdefault(pos, [])
                    for label in base_labels:
                        cur.append(label)

                    if len(item.spans) > 1 and has_link:
                        cur.append('ref')
                        if is_self_link:
                            cur.append('self-link')

                    if pos == span.end:
                        break
                    pos = self.doc.get_next_pos(pos)
                    # Handle the case of a space
                    if len(pos) == 2 or (len(pos) == 3 and pos[2] == 0):
                        cur = ans.setdefault((pos[0], pos[1], -1), [])
                        for label in base_labels:
                            cur.append(label)
                        if len(item.spans) > 1 and has_link:
                            cur.append('ref')

        # Now do disagreement colours.
        for item, count in self.disagreements:
            # Get the standard color for this item based on its label
            base_labels = []
            if self.config.annotation_type == 'categorical':
                for key in item.labels:
                    if key in self.config.labels:
                        base_labels.append("compare-{}-{}".format(count, key))
                    else:
                        base_labels.append("compare-label-{}-{}".format(count, key))
            
            has_link = False
            for span in item.spans:
                if span == linking_pos:
                    has_link = True

            ref_label = "compare-ref-{}-{}".format(has_link, count)
            max_span = max(item.spans)
            for span in item.spans:
                pos = span.start
                while True:
                    cur = ans.setdefault(pos, [])
                    for label in base_labels:
                        cur.append(label)
                        # TODO: Record the span too

                    if len(item.spans) > 1:
                        if span == max_span:
                            cur.append(ref_label +"-last")
                        else:
                            cur.append(ref_label +"-earlier")

                    if pos == span.end:
                        break
                    pos = self.doc.get_next_pos(pos)
                    # Handle the case of a space
                    if len(pos) == 2 or (len(pos) == 3 and pos[2] == 0):
                        cur = ans.setdefault((pos[0], pos[1], -1), [])
                        for label in base_labels:
                            cur.append(label)
                        if len(item.spans) > 1 and has_link:
                            if span == max_span:
                                cur.append(ref_label +"-last")
                            else:
                                cur.append(ref_label +"-earlier")

        return ans

    def next_match(self, span, text, reverse=False):
        return self.doc.next_match(span, text, reverse)

    def get_item_with_spans(self, spans, any_present=False):
        items = []
        for item in self.annotations:
            match = 0
            for span in item.spans:
                if span in spans:
                    match += 1
            rev_match = 0
            for span in spans:
                if span in item.spans:
                    rev_match += 1
            if len(item.spans) == len(spans) == match == rev_match:
                items.append(item)
            elif any_present and match > 0:
                items.append(item)
        return items

    def modify_annotation(self, spans, label=None):
        # TODO: switch link to be like the old style
        to_edit = self.get_item_with_spans(spans)
        if len(to_edit) == 0:
            # No item with these spans exists, create it
            nspans = [Span(self.config.annotation, self.doc, s) for s in spans]
            item = Item(self.doc, nspans, label)
            self.annotations.append(item)
        else:
            for item in to_edit:
                # Modify existing item
                if label is None:
                    if len(item.labels) == 0:
                        self.annotations.remove(item)
                elif label in item.labels:
                    item.labels.remove(label)
                    if len(item.labels) == 0:
                        self.annotations.remove(item)
                else:
                    item.labels.add(label)

    def remove_annotation(self, spans):
        permissive = self.config.annotation_type == 'link'
        for item in self.get_item_with_spans(spans, permissive):
            self.annotations.remove(item)

    def write_out(self, filename=None):
        out_filename = self.output_file
        if filename is not None:
            out_filename = filename
        out = open(out_filename, 'w')
        for item in self.annotations:
            print(str(item), file=out)
        out.close()




from __future__ import print_function

import _curses
import logging
import curses
import re
import sys

from .config import *
from .data import Span, span_compare_ge, span_compare_le

class View(object):
    def __init__(self, window, cursor, linking_pos, datum, my_config, cnum, total_num, prev_view=None):
        self.window = window
        self.cursor = cursor
        self.linking_pos = linking_pos
        self.datum = datum
        self.top = max(0, cursor.start[0] - self.window.getmaxyx()[0] - 10)
        self.show_help = False
        self.show_progress = False
        self.progress = "file {} / {}".format(cnum + 1, total_num)
        self.show_legend = False
        self.show_current_mark = False
        self.legend = []
        self.config = my_config
        self.line_numbers = False
        if prev_view is not None:
            self.show_help = prev_view.show_help
            self.show_progress = prev_view.show_progress
            self.show_legend = prev_view.show_legend
            self.line_numbers = prev_view.line_numbers

        self.last_moved_pos = cursor

        if self.config.annotation_type == 'categorical':
            for label, info in self.config.labels.items():
                self.legend.append(("{} {}".format(label, "+".join(info[0])), label))

        if self.config.args.prevent_self_links and self.cursor == self.linking_pos:
            # TODO: In this case move the linking pos along one step
            pass

    def instructions(self):
        shared = [
            "Misc      | ] [                   | save and go to next ] / previous [ file",
            "          | q Q                   | quit with or without saving            ",
            "          | s                     | save the current file                  ",
        ]
        if self.config.annotation_type == 'link':
            return [
                "Keybindings - to hide/show this, type 'h'",
                "Colours are underline:selected, green:linking, blue:linked, yellow:has_link",
                "Type      | Key                   | Affect                             ",
                "----------|-----------------------|------------------------------------",
                "Move      | j or LEFT             | move selected item to the left     ",
                "          | J or [SHIFT + LEFT]   | move linking item to the left      ",
                "          | i or UP               | move selected item up a line       ",
                "          | I or [SHIFT + UP]     | move linking item up a line        ",
                "          | o or DOWN             | move selected item down a line     ",
                "          | O or [SHIFT + DOWN]   | move linking item down a line      ",
                "          | ; or RIGHT            | move selected item to the right    ",
                "          | : or [SHIFT + RIGHT]  | move linking item to the right     ",
                "Annotate  | d                     | create a link and move down / right",
                "          | D                     | create a link                      ",
                "          | u                     | undo all annotations for this item ",
            ] + shared
        else:
            return [
                "Keybindings - to hide/show this, type 'h'",
                "Underline is current item, colours are labels",
                "Type      | Key                   | Affect                           ",
                "----------|-----------------------|----------------------------------",
                "Move      | j or LEFT             | move to the left                 ",
                "          | J or [SHIFT + LEFT]   | go to the start of the line      ",
                "          | i or UP               | move up a line                   ",
                "          | I or [SHIFT + UP]     | go to first line                 ",
                "          | o or DOWN             | move down a line                 ",
                "          | O or [SHIFT + DOWN]   | go to last line                  ",
                "          | ; or RIGHT            | move to the right                ",
                "          | : or [SHIFT + RIGHT]  | go to the end of the line        ",
                "Span Edit | m   /                 | extend left or right             ",
                "          | k   l                 | contract left or right           ",
                "Annotate  | SPACE then a, s, or d | [un]mark this item as a, s, or d ",
                "          | u                     | undo annotation on this item     ",
            ] + shared

    def toggle_help(self):
        self.show_help = not self.show_help
    def toggle_progress(self):
        self.show_progress = not self.show_progress
    def toggle_legend(self):
        self.show_legend = not self.show_legend
    def toggle_current_mark(self):
        self.show_current_mark = not self.show_current_mark

    def shift_view(self, down=False):
        self.last_moved_pos = None
        if down: self.top += 10
        else: self.top -= 10

    def _check_move_allowed(self, move_link, new_pos):
        if self.linking_pos is None:
            return True
        if self.config.args.prevent_forward_links:
            if move_link and self.cursor > new_pos:
                return False
            if (not move_link) and self.linking_pos < new_pos:
                return False
        if self.config.args.prevent_self_links:
            if move_link and self.cursor == new_pos:
                return False
            if (not move_link) and self.linking_pos == new_pos:
                return False
        return True

    def move(self, direction, distance, maxjump=False, move_link=False):
        # TODO: if self links are prevented, but forward links are allowed,
        # then skip over when moving.
        mover = self.cursor
        if move_link:
            mover = self.linking_pos
###            logging.debug("Moving linking pos")
###        logging.debug("Move {} {} {}".format(self.cursor, direction, distance))
        new_pos = mover.edited(direction, 'move', distance, maxjump)
###        logging.debug("Moving {} to {}".format(self.cursor, new_pos))
        if self._check_move_allowed(move_link, new_pos):
            if move_link:
                self.linking_pos = new_pos
            else:
                self.cursor = new_pos
            self.last_moved_pos = new_pos
        elif self.linking_pos is not None:
            # Move as far as possible
            if self.config.args.prevent_self_links:
                if move_link:
                    self.linking_pos = self.cursor.edited('next', 'move', 1, False)
                    self.last_moved_pos = self.linking_pos
                else:
                    self.cursor = self.linking_pos.edited('previous', 'move', 1, False)
                    self.last_moved_pos = self.cursor
            else:
                if move_link:
                    self.linking_pos = self.cursor
                else:
                    self.cursor = self.linking_pos
                self.last_moved_pos = self.cursor

    def search(self, query, direction, count, maxjump=False, move_link=False):
        logging.debug("Search {} {} {} {} {}".format(query, direction, count, maxjump, move_link))
        new_pos = None
        if query is None:
            if len(self.datum.disagreements) == 0:
                new_pos = self.datum.get_next_unannotated(self.cursor, self.linking_pos, direction, move_link)
                if new_pos == self.linking_pos or new_pos is None:
                    new_pos = self.datum.get_next_self_link(self.cursor, self.linking_pos, direction, move_link)
            else:
                new_pos = self.datum.get_next_disagreement(self.cursor, self.linking_pos, direction, move_link)
        else:
            mover = self.cursor
            if move_link:
                mover = self.linking_pos
            new_pos = mover.search(query, direction, count, maxjump)

        if new_pos is not None:
            if self._check_move_allowed(move_link, new_pos):
                if move_link:
                    self.linking_pos = new_pos
                else:
                    self.cursor = new_pos
                self.last_moved_pos = new_pos

    def adjust(self, direction, distance, change, maxjump, link):
        changer = self.cursor
        if link and self.linking_pos is not None:
            changer = self.linking_pos

        new_pos = changer.edited(direction, change, distance, maxjump)
        if self._check_move_allowed(link, new_pos):
            if link:
                self.linking_pos = new_pos
            else:
                self.cursor = new_pos
    
    def put_cursor_beside_link(self):
        self.cursor = self.linking_pos.edited('previous')

    def marking_to_color(self, marking):
        name = DEFAULT_COLOR
        modifier = curses.A_BOLD
        has_link = False
        has_ref = False
        has_self_link = False
        for mark in marking:
            if mark == 'cursor':
                modifier += curses.A_UNDERLINE
            elif mark == 'link':
                has_link = True
            elif mark == 'ref':
                has_ref = True
            elif mark == 'self-link':
                has_self_link = True
            elif mark == 'linked':
                name = IS_LINKED_COLOR
            elif mark in self.config.labels:
                if name != DEFAULT_COLOR:
                    name = OVERLAP_COLOR
                else:
                    name = self.config.get_color_for_label(mark)
            elif mark.startswith("compare-"):
                if 'ref' in mark:
                    count = int(mark.split("-")[-2])
                    if 'True' in mark and 'last' not in mark:
                        # First, cases where this is related to the current linking line
                        if name == DEFAULT_COLOR or name == COMPARE_DISAGREE_COLOR:
                            if count == 0:
                                name = REF_COLOR
                            else:
                                name = COMPARE_REF_COLOR
                    elif count > 0 and 'last' in mark:
                        # If unrelated, but there is a disagreement, indicate it
                        if name == DEFAULT_COLOR:
                            name = COMPARE_DISAGREE_COLOR
                else:
                    count = int(mark.split("-")[-2])
                    if name == DEFAULT_COLOR:
                        if count == 0:
                            key = mark.split("-")[-1]
                            if key in self.config.labels:
                                name = self.config.get_color_for_label(key)
                        else:
                            name = COMPARE_DISAGREE_COLOR
###                    key = mark.split("-")[-1]
###                    if key in self.config.labels:
###                        name = self.config.get_color_for_label(mark)
        # Override cases
        if has_link:
            if has_ref:
                if has_self_link:
                    name = SELF_LINK_COLOR
                else:
                    name = IS_LINKED_COLOR
            else:
                name = LINK_COLOR
        elif has_ref:
            name = REF_COLOR

        return curses.color_pair(name) + modifier

    def do_contents(self, height, width, markings, number_width, trial=False):
        # For linked items, colour them to indicate it
        # For labels, colour them always, and add beginning / end
        # For freeform text, include it at the bottom

        # Row and column indicate the position on the screen, while line and
        # token indicate the position in the text.
        first_span = None
        last_span = None
        row = -1
        for line_no, line in enumerate(self.datum.doc.tokens):
            # If this line is above the top of what we are shwoing, skip it
            if line_no < self.top:
                continue
            if row >= height:
                break
            if first_span is None:
                first_span = Span('character', self.datum.doc, (line_no, 0, 0))

            # Set
            row += 1
            column = number_width
            if (not trial) and column > 0:
                self.window.addstr(row, 0, str(line_no), curses.color_pair(LINE_NUMBER_COLOR))
            for token_no, token in enumerate(line):
                # Check if we are going on to the next line and adjust
                # accordingly.
                space_before = 1 if column > number_width else 0
                wide_token = False
                if column + len(token) + space_before > width:
                    if token_no != 0:
                        column = 0
                        row += 1
                        space_before = 0
                    else:
                        wide_token = True

                # If this takes us off the screen, stop
                if row >= height:
                    break

                end_pos = len(token) - 1
                if wide_token:
                    end_pos = width - column - space_before - 1
                last_span = Span('character', self.datum.doc, (line_no, token_no, end_pos))

                for char_no, char in enumerate(token):
                    if column >= width:
                        column = 0
                        row += 1
                        if row >= height:
                            break

                    # Allow multiple layers of color, with the more specific
                    # domainating
                    if space_before > 0:
                        if not trial:
                            mark = []
                            if () in markings:
                                mark = markings[()]
                            if (line_no,) in markings:
                                mark = markings[(line_no,)]
                            if (line_no, token_no, -1) in markings:
                                mark = markings[line_no, token_no, -1]
                            color = self.marking_to_color(mark)
                            self.window.addstr(row, column, ' ', color)
                        column += 1
                        space_before = 0

                    color = self.marking_to_color([])

                    if not trial:
                        mark = []
                        if () in markings:
                            mark = markings[()]
                        if (line_no,) in markings:
                            mark = markings[(line_no,)]
                        if (line_no, token_no) in markings:
                            mark = markings[line_no, token_no]
                        if (line_no, token_no, char_no) in markings:
                            mark = markings[line_no, token_no, char_no]
                        color = self.marking_to_color(mark)
                        try:
                            self.window.addstr(row, column, char, color)
                        except _curses.error as e:
                            logging.warn("Error caught in drawing extra lines 2")
                    column += 1

                if row >= height:
                    break

        # Tracks if we can see where we moved to
        seen_last_moved_pos = True
        if first_span is not None and last_span is not None:
            if self.last_moved_pos is not None:
                cmp_first = self.last_moved_pos.compare(first_span)
                cmp_last = self.last_moved_pos.compare(last_span)
                seen_last_moved_pos = cmp_first in span_compare_ge and \
                        cmp_last in span_compare_le
        return seen_last_moved_pos


    def render(self, current_search, current_typing):
        height, width = self.window.getmaxyx()

###        logging.debug("Render with '{}' and '{}'".format(current_search, current_typing))
        if self.show_help:
            if height < 20 or width < 80:
                raise Exception("Window too small - must be at least 80x20")
            self.render_help()
            return

        # Get content for extra lines
        extra_text_lines = []
        if self.show_progress: extra_text_lines.append(self.progress)
        if self.show_legend:
            cur = []
            length = 0
            for val, colour in self.legend:
                if length + 3 + len(val) > width:
                    extra_text_lines.append(cur)
                    cur = []
                    length = 0
                if length > 0:
                    length += 3
                length += len(val)
                cur.append((val, colour))
            extra_text_lines.append(cur)
        if len(current_search) > 0: extra_text_lines.append(current_search)
        if len(current_typing) > 0: extra_text_lines.append(current_typing)

        # Get colors for content
        markings = self.datum.get_all_markings(self.cursor, self.linking_pos)
###        logging.debug("markings: {}".format(markings))
        if self.show_current_mark:
            for key in markings:
                logging.debug("marking: {}: {}".format(key, markings[key]))
                marks = markings[key]
                if 'cursor' in marks:
                    for mark in marks:
                        if mark in self.config.labels:
                            extra_text_lines.append("Current: "+ mark)
                        elif 'compare-' in mark and 'ref' not in mark:
                            parts = mark.split("-")
                            count = len(self.datum.other_annotations) - int(parts[-2])
                            label = parts[-1]
                            extra_text_lines.append("{} marked as {}".format(count, label))

        # First, plan instructions
        main_height = height
        space_needed = len(extra_text_lines)
        if space_needed > 0:
            space_needed += 1
        elif len(self.datum.other_annotations) > 0:
            pass

        if height >= space_needed:
            main_height = main_height - space_needed

        # Shift the top up if necessary
        if self.config.annotation != 'document':
            if self.last_moved_pos is not None:
                if self.top > self.last_moved_pos.start[0]:
                    self.top = self.last_moved_pos.start[0]
                if self.top < self.last_moved_pos.start[0] - main_height:
                    self.top = self.last_moved_pos.start[0] - main_height

        # Work out width
        main_width = width
        number_width = 0
        if self.line_numbers:
            count = 1
            lines = len(self.datum.doc.lines)
            while lines > 1:
                count += 1
                lines /= 10
            number_width = count
            main_width -= number_width

        # Do dry runs, shifting top down until the position is visible
        while not self.do_contents(main_height, main_width, markings, number_width, True):
            self.top += 1

        # Next, draw contents
        self.do_contents(main_height, main_width, markings, number_width)

        if main_height < height:
            if len(extra_text_lines) > 0:
                row = main_height
                self.window.addstr(row, 0, "-" * width)
                row += 1

                color = curses.color_pair(HELP_COLOR) + curses.A_BOLD

                # Last, draw the text being typed
                for content in extra_text_lines:
                    try:
                        if isinstance(content, list):
                            position = 0
                            for text, tcolor in content:
                                if position != 0:
                                    self.window.addstr(row, position, " | ", color)
                                    position += 3
                                name = self.config.get_color_for_label(tcolor)
                                tcolor = curses.color_pair(name) + curses.A_BOLD
                                self.window.addstr(row, position, text, tcolor)
                                position += len(text)
                        else:
                            text = content
                            fmt = "{:<"+ str(width) +"}"
                            self.window.addstr(row, 0, fmt.format(text), color)
                        row += 1
                    except _curses.error as e:
                        logging.warn("Error caught in drawing extra lines")

        self.window.refresh()

    def render_help(self):
        self.window.clear()
        height, width = self.window.getmaxyx()

        row = 0
        color = curses.color_pair(HELP_COLOR) + curses.A_BOLD
        for line in self.instructions():
            fmt = "{:<"+ str(width) +"}"
            self.window.addstr(row, 0, fmt.format(line), color)
            row += 1

        self.window.refresh()

    def render_edgecase(self, at_end):
        self.window.clear()
        height, width = self.window.getmaxyx()
        pos = int(height / 2)
        edge = 'end' if at_end else 'start'
        line0 = "At "+ edge +" in the files."
        dir_key = '[' if at_end else ']'
        line1 = "Type 'q' to quit, or '"+ dir_key+ "' to go back."
        self.window.addstr(pos, 0, line0, curses.color_pair(HELP_COLOR) + curses.A_BOLD)
        self.window.addstr(pos + 1, 0, line1, curses.color_pair(DEFAULT_COLOR) + curses.A_BOLD)
        self.window.refresh()




from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="slate-nlp",
    version="1.0.1",
    author="Jonathan K. Kummerfeld",
    author_email="jkk@berkeley.edu",
    description="A terminal-based text annotation tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://jkk.name/slate/",
    packages=find_packages(),
    keywords="nlp annotation labeling natural-language-processing text-annotation",
    python_requires='>=2.6, !=3.0.*, !=3.1.*, !=3.2.*, <4',
    entry_points={
        'console_scripts': [
            'slate = slate.annotate:main',
        ],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "License :: OSI Approved :: ISC License (ISCL)",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)



from __future__ import print_function

import argparse
import curses
import datetime
import logging
import string
import sys

from .data import *
from .config import *
from .view import *

class Annotator(object):
    def __init__(self, config, filenames, current_mode, args):
        self.current_mode = current_mode
        self.current_num = None
        self.search_term = ''
        self.partial_typing = ''
        self.cfilename = -1
        self.filename = None
        self.filenames = filenames
        self.datum = None
        self.view = None
        self.window = None
        self.config = config
        self.args = args
        self.action_to_function = {
            'delete-query-char': self.delete_typing_char,
            'leave-query-mode': self.leave_typing_mode,
            'enter-query-mode': self.enter_typing_mode,
            'clear-query': self.clear_query,
            'add-to-query': self.add_to_typing,
            'delete-label-char': self.delete_typing_char,
            'assign-text-label': self.assign_text,
            'enter-label-mode': self.enter_typing_mode,
            'add-to-label': self.add_to_typing,
            'toggle-line-numbers': self.toggle_line_numbers,
            'move-up': self.move,
            'move-down': self.move,
            'move-left': self.move,
            'move-right': self.move,
            'move-link-up': self.move,
            'move-link-down': self.move,
            'move-link-left': self.move,
            'move-link-right': self.move,
            'jump-up': self.move,
            'jump-down': self.move,
            'jump-left': self.move,
            'jump-right': self.move,
            'extend-up': self.change_span,
            'extend-down': self.change_span,
            'extend-left': self.change_span,
            'extend-right': self.change_span,
            'contract-up': self.change_span,
            'contract-down': self.change_span,
            'contract-left': self.change_span,
            'contract-right': self.change_span,
            'extend-link-up': self.change_span,
            'extend-link-down': self.change_span,
            'extend-link-left': self.change_span,
            'extend-link-right': self.change_span,
            'contract-link-up': self.change_span,
            'contract-link-down': self.change_span,
            'contract-link-left': self.change_span,
            'contract-link-right': self.change_span,
            'search-previous': self.search,
            'search-next': self.search,
            'search-link-previous': self.search,
            'search-link-next': self.search,
            'page-up': self.shift_view,
            'page-down': self.shift_view,
            'toggle-help': self.modify_display,
            'toggle-progress': self.modify_display,
            'toggle-legend': self.modify_display,
            'toggle-current-mark': self.modify_display,
            'next-file': self.change_file,
            'previous-file': self.change_file,
            'quit': self.save_or_quit,
            'save-and-quit': self.save_or_quit,
            'save': self.save_or_quit,
            'create-link': self.create_link,
            'create-link-and-move': self.create_link,
            'edit-annotation': self.edit_annotation,
            'remove-annotation': self.remove_annotation,
            'update-num': self.update_number,
        }

    def move(self, user_input, action):
        if self.current_mode[-1] == 'no_file':
            return

        direction = action.split('-')[-1]
        jump = 'jump' in action
        link = 'link' in action

        num = 1
        if self.current_num == 0:
            jump = True
            self.current_num = None
        elif self.current_num is not None:
            num = self.current_num
            self.current_num = None

        self.view.move(direction, num, jump, link)

    def toggle_line_numbers(self, user_input, action):
        self.view.line_numbers = not self.view.line_numbers

    def change_span(self, user_input, action):
        if self.current_mode[-1] == 'no_file':
            return

        change = action.split('-')[0]
        direction = action.split('-')[-1]
        # TODO: Support these adjustments to the linking_pos too. This requires
        # edits to view and also to config
        link = 'link' in action

        num = 1
        jump = False
        if self.current_num == 0:
            jump = True
            self.current_num = None
        elif self.current_num is not None:
            num = self.current_num
            self.current_num = None

        self.view.adjust(direction, num, change, jump, link)

    def delete_typing_char(self, user_input, action):
        if self.current_mode[-1] == 'no_file':
            return

        if self.current_mode[-1] == 'write_query':
            self.search_term = self.search_term[:-1]
        else:
            self.partial_typing = self.partial_typing[:-1]

    def leave_typing_mode(self, user_input, action):
        if self.current_mode[-1] == 'no_file':
            return

        if len(self.current_mode) > 1:
            self.current_mode.pop()

    def assign_text(self, user_input, action):
        if self.current_mode[-1] == 'no_file':
            return

        if len(self.current_mode) > 1:
            self.current_mode.pop()
            self.datum.modify_annotation([self.view.cursor], self.partial_typing)
            self.partial_typing = ''

    def enter_typing_mode(self, user_input, action):
        if self.current_mode[-1] == 'no_file':
            return

        if 'query' in action:
            self.current_mode.append('write_query')
        else:
            self.current_mode.append('write_label')
            self.partial_typing = ''

    def clear_query(self, user_input, action):
        if self.current_mode[-1] == 'no_file':
            return

        self.search_term = ''

    def add_to_typing(self, user_input, action):
        if self.current_mode[-1] == 'no_file':
            return

        char = user_input[0]
        if user_input[0] == 'SPACE':
            char = ' '
        if self.current_mode[-1] == 'write_query':
            self.search_term += char
        else:
            self.partial_typing += char

    def change_file(self, user_input, action):
        if self.current_mode[-1] != 'no_file':
            self.save_or_quit(None, 'save')

        direction = 1 if 'next' in action else -1
        if self.current_mode[-1] == 'no_file':
            if (self.cfilename < 0) == (direction > 0):
                self.current_mode.pop()
                self.cfilename += direction
        elif 0 <= self.cfilename + direction < len(self.filenames):
            self.cfilename += direction
            self.filename, start_pos, output_file, annotation_files = \
                    self.filenames[self.cfilename]
            self.datum = Datum(self.filename, self.config, output_file, annotation_files)
            self.get_view(self.datum, self.config, self.cfilename, len(self.filenames), start_pos, self.view)
        elif self.current_mode != 'no_file':
            self.cfilename += direction
            self.current_mode.append('no_file')

    def modify_display(self, user_input, action):
        if self.current_mode[-1] == 'no_file':
            return

        if 'help' in action:
            self.view.toggle_help()
        elif 'progress' in action:
            self.view.toggle_progress()
        elif 'legend' in action:
            self.view.toggle_legend()
        elif 'current-mark' in action:
            self.view.toggle_current_mark()

    def shift_view(self, user_input, action):
        if self.current_mode[-1] == 'no_file':
            return

        if 'up' in action:
            self.view.shift_view()
        else:
            self.view.shift_view(True)

    def update_number(self, user_input, action):
        if self.current_mode[-1] == 'no_file':
            return

        num = int(user_input[0])
        if self.current_num is None:
            self.current_num = 0
        else:
            self.current_num *= 10
        self.current_num += num

    def remove_annotation(self, user_input, action):
        if self.current_mode[-1] == 'no_file':
            return

        if self.current_mode[-1] != 'read':
            spans = [self.view.cursor]
            if self.current_mode[-1] == 'link':
                spans = [self.view.linking_pos]
            self.datum.remove_annotation(spans)

    def edit_annotation(self, user_input, action):
        if self.current_mode[-1] == 'no_file':
            return

        if self.current_mode[-1] == 'category':
            label = self.config.get_label_for_input(user_input)
            self.datum.modify_annotation([self.view.cursor], label)

    def create_link(self, user_input, action):
        if self.current_mode[-1] == 'no_file':
            return

        self.datum.modify_annotation([self.view.cursor, self.view.linking_pos])
        if 'and-move' in action:
            if self.config.annotation == 'line':
                self.view.move('down', 1, False, True)
                self.view.put_cursor_beside_link()
            else:
                self.view.move('right', 1, False, True)
                self.view.put_cursor_beside_link()
            self.view.must_show_linking_pos = True

    def save_or_quit(self, user_input, action):
        if 'save' in action:
            if self.current_mode[-1] != 'read':
                self.datum.write_out()

            # TODO: Save both cursor and linking pos
            if 0 <= self.cfilename < len(self.filenames):
                cur = self.filenames[self.cfilename]
                pos = self.view.cursor
                if self.config.annotation_type == 'link':
                    pos = self.view.linking_pos
                self.filenames[self.cfilename] = (cur[0], pos, cur[2], cur[3])

        if 'quit' in action:
            if 'save' not in action:
                # TODO: Have an 'are you sure?' step
                pass
            return 'quit'

    def search(self, user_input, action):
        if self.current_mode[-1] == 'no_file':
            return

        direction = action.split('-')[-1]
        jump = False
        link = 'link' in action

        num = 1
        if self.current_num == 0:
            jump = True
            self.current_num = None
        elif self.current_num is not None:
            num = self.current_num
            self.current_num = None

        if len(self.search_term) > 0:
            self.view.search(self.search_term, direction, num, jump, link)
        else:
            # Used when comparing files to go to the next/previous annotation
            # Or when not comparing files to go to the next unannotated thing
            # When there is nothing unannotated, it jumps to the next self-linked item
            self.view.search(None, direction, num, jump, link)


    def input_to_symbol(self, num):
        if num in key_to_symbol:
            return key_to_symbol[num]
        else:
            return "UNKNOWN"

    def get_view(self, config, file_num, total_files, position, prev_view=None):
        cursor = position
        link = position if self.config.annotation_type == 'link' else None
        self.view = View(self.window, cursor, link, self.datum, self.config, file_num, total_files, prev_view)

    def annotate(self, window_in):
        self.window = window_in

        # Set color combinations
        curses.use_default_colors()
        for num, fore, back in COLORS:
            curses.init_pair(num, fore, back)

        # No blinking cursor
        curses.curs_set(0)

        self.cfilename = 0
        self.filename, start_pos, output_file, annotation_files = self.filenames[self.cfilename]
        self.datum = Datum(self.filename, self.config, output_file, annotation_files)
        self.get_view(self.config, self.cfilename, len(self.filenames), start_pos)
        if not self.args.hide_help:
            self.view.toggle_help()

        last_num = None
        at_end = None
        nsteps = 0
        user_input = []
        while True:
            # Draw screen
            if self.current_mode[-1] == 'no_file':
                self.view.render_edgecase(self.cfilename >= 0)
            else:
                # Set current search term appearance
                tmp_term = self.search_term
                if self.current_mode[-1] == 'write_query':
                    tmp_term = '\\'+ tmp_term

                self.view.render(tmp_term, self.partial_typing)
            self.view.must_show_linking_pos = False

            # Get input
            ch = self.window.getch()
            next_user_input = self.input_to_symbol(ch)
            logging.debug("Input {} converted to {} in mode {}".format(ch, next_user_input, self.current_mode))
            user_input.append(next_user_input)
            tuser_input = tuple(user_input)
            if (self.current_mode[-1], tuser_input) not in self.config.valid_prefixes:
                if (None, tuser_input) not in self.config.valid_prefixes:
                    if (self.current_mode[-1], tuser_input) not in self.config.input_to_action:
                        if (None, tuser_input) not in self.config.input_to_action:
                            user_input = [next_user_input]
                            tuser_input = (next_user_input,)
            nsteps += 1
            if nsteps % 100 == 0 and self.current_mode[-1] == 'category':
                self.datum.write_out()

            # Determine what to do for the input
            action = None
            function = None
            if (self.current_mode[-1], tuser_input) in self.config.input_to_action:
                action = self.config.input_to_action[self.current_mode[-1], tuser_input]
                if action in self.action_to_function:
                    function = self.action_to_function[action]
            elif (None, tuser_input) in self.config.input_to_action:
                action = self.config.input_to_action[None, tuser_input]
                if action in self.action_to_function:
                    function = self.action_to_function[action]
            logging.debug("{} {} -> {} {}".format(self.current_mode, tuser_input, action, function))

            # Do it!
            if function is not None:
                outcome = function(tuser_input, action)
                user_input = []
                if outcome == 'quit':
                    break

            # Clear the screen in preparation for rendering it again
            self.window.clear()

        # Write out information for continuing annotation later
        out_filename = self.args.log_prefix + '.todo'
        out = open(out_filename, "w")
        for fname, start_pos, output_file, annotation_files in self.filenames:
            parts = [
                fname,
                output_file,
                str(start_pos), # TODO - simplified value here
                ' '.join(annotation_files)
            ]
            print(" ".join(parts), file=out)
        out.close()

def ext_annotate(window_in, annotator):
    annotator.annotate(window_in)

def main():
    stime = datetime.datetime.now().strftime('%Y-%m-%d.%H-%M-%S')

    parser = argparse.ArgumentParser(
            description='A tool for annotating text data.',
            fromfile_prefix_chars='@')
    parser.add_argument('data', nargs="*",
            help='Files to be annotated')
    parser.add_argument('-d', '--data-list', nargs="+",
            help='Files containing lists of files to be annotated')

    parser.add_argument('-t', '--ann-type',
            choices=['categorical', 'link'],
            default='categorical',
            help='The type of annotation being done.')
    parser.add_argument('-s', '--ann-scope',
            choices=['character', 'token', 'line', 'document'],
            default='token',
            help='The scope of annotation being done.')
    parser.add_argument('-c', '--config-file',
            help='A file containing configuration information.')
    parser.add_argument('-l', '--log-prefix',
            default="annotation_log."+ stime,
            help='Prefix for logging files')
    parser.add_argument('-ld', '--log-debug',
            action='store_true',
            help='Provide detailed logging.')

    parser.add_argument('-hh', '--hide-help',
            action='store_true',
            help='Do not show help on startup.')
    parser.add_argument('-r', '--readonly',
            action='store_true',
            help='Do not allow changes or save annotations.')
    parser.add_argument('-o', '--overwrite',
            default=False,
            action='store_true',
            help='If they exist already, read abd overwrite output files.')

    parser.add_argument('-ps', '--prevent-self-links',
            default=False,
            action='store_true',
            help='Prevent an item from being linked to itself.')
    parser.add_argument('-pf', '--prevent-forward-links',
            default=False,
            action='store_true',
            help='Prevent a link from an item to one after it.')

    parser.add_argument('--do-not-show-linked',
            default=False,
            action='store_true',
            help='Do not have a special color to indicate any linked token.')
    parser.add_argument('--alternate-comparisons',
            default=False,
            action='store_true',
            help='Activate alternative way of showing different annotations '
            '(one colour per set of markings, rather than counts).')

    args = parser.parse_args()

    if len(args.data) == 0 and args.data_list is None:
        parser.error("No filenames or data lists provided")

    # Set up logging
    logging_level = logging.DEBUG if args.log_debug else logging.INFO
    logging.basicConfig(filename=args.log_prefix + '.log', level=logging_level)
    logging.info("Executed with: {}".format(' '.join(sys.argv)))
    logging.info("Arguments interpreted as: {}".format(args))
    if logging_level == logging.INFO:
        sys.tracebacklimit = 1

    ### Process configuration
    config = None
    if args.config_file is not None:
        config = Config(args)
    else:
        config = Config(args, 
            {
                'label:a': (('SPACE', 'a'), 'green'),
                'label:s': (('SPACE', 's'), 'blue'),
                'label:d': (('SPACE', 'd'), 'magenta'),
                'label:v': (('SPACE', 'v'), 'red'),
            }
        )

    file_info = args.data
    if args.data_list is not None:
        for filename in args.data_list:
            if len(glob.glob(filename)) == 0:
                raise Exception("Cannot open / find '{}'".format(filename))
            for line in open(filename):
                file_info.append(line.strip())

    filenames = process_fileinfo(file_info, config)
    if len(filenames) == 0:
        print("Found no files")
        sys.exit(0)

    config_out = open(args.log_prefix + '.config', 'w')
    print(config, file=config_out)
    config_out.close()

    # Set the current mode
    current_mode = []
    if args.readonly:
        current_mode.append('read')
    elif args.ann_type == 'categorical':
        current_mode.append('category')
    elif args.ann_type == 'link':
        current_mode.append('link')

    ### Start interface
    annotator = Annotator(config, filenames, current_mode, args)
    curses.wrapper(ext_annotate, annotator)
